import time
from typing import Literal

import attr
from luna3.events.events import EventsApi
from luna3.image_store.image_store import StoreApi
from luna3.sender.wsclient import LunaWSClient

from .protocols import AggregatedEventT, HTTPClient


class NotInitialized(Exception):
    """`Not initialized` internal exception"""


@attr.dataclass(slots=True)
class EventMetadata:
    """Event metadata for event enrichment by data from luna-video-agent"""

    accountId: str
    eventCreateTime: str
    eventEndTime: str
    location: dict[str, str]
    streamId: str
    source: str | None
    meta: dict | None


@attr.dataclass(slots=True)
class LunaApiExternalParams:
    """Luna api external params for correct external image links creation"""

    origin: str
    apiVersion: int


@attr.dataclass(slots=True)
class LunaContext:
    """Luna context"""

    eventMetadata: EventMetadata
    requestId: str
    lisBucket: str
    eventsClient: EventsApi | None = None
    senderClient: LunaWSClient | None = None
    httpClient: HTTPClient | None = None


class LunaWraps:
    """Warp for luna services client, metadata and so on"""

    # luna api external params, must be initialized by luna-agent externally
    _lunaApiExternalParams: LunaApiExternalParams | None = None
    # luna events client, must be initialized by luna-agent externally
    _eventsClient: EventsApi | None = None
    # luna image store client, must be initialized by luna-agent externally
    _lunaImageStoreClient: StoreApi | None = None
    # luna sender client, must be initialized by luna-agent externally
    _senderClient: LunaWSClient | None = None
    # http client, must be initialized by luna-agent externally
    _httpClient: HTTPClient | None = None
    # request id, must be unique for each event and must be the same for all requests concerning one event
    _requestId: str | None = None

    def __init__(self, eventId: str, eventMetadata: EventMetadata, **kwargs):
        self.eventId = eventId
        self.eventMetadata = eventMetadata
        self.lunaContext = LunaContext(
            eventMetadata=eventMetadata,
            requestId=self.requestId,
            eventsClient=self.eventsClient,
            senderClient=self.senderClient,
            httpClient=self.httpClient,
            lisBucket=self.lisBucket,
        )
        self._requestId = None

    @property
    def requestId(self) -> str:
        """Request id lazy getter"""
        if self._requestId is None:
            self._requestId = f"{int(time.time())},{self.eventId}"
        return self._requestId

    @classmethod
    async def initialize(
        cls,
        lunaEventsClient: EventsApi,
        lunaImageStoreClient: StoreApi,
        lunaSenderClient: LunaWSClient,
        lunaApiExternalParams: LunaApiExternalParams,
        httpClient: HTTPClient,
        lisBucket: str,
    ):
        """Initialization of luna clients and so on"""
        cls._eventsClient = lunaEventsClient
        cls._lunaImageStoreClient = lunaImageStoreClient
        cls._lunaSenderClient = lunaSenderClient
        cls._lunaApiExternalParams = lunaApiExternalParams
        cls._httpClient = httpClient
        cls.lisBucket = lisBucket

    @property
    def lunaApiExternalParams(self) -> LunaApiExternalParams:
        """Luna api external params lazy getter"""
        if self._lunaApiExternalParams is None:
            raise NotInitialized
        return self._lunaApiExternalParams

    @property
    def eventsClient(self) -> EventsApi:
        """Luna event client lazy getter"""
        if self._eventsClient is None:
            raise NotInitialized
        return self._eventsClient

    @property
    def lunaImageStoreClient(self) -> StoreApi:
        """Luna image store client lazy getter"""
        if self._lunaImageStoreClient is None:
            raise NotInitialized
        return self._lunaImageStoreClient

    @property
    def senderClient(self) -> LunaWSClient:
        """Luna sender client lazy getter"""
        if self._lunaSenderClient is None:
            raise NotInitialized
        return self._lunaSenderClient

    @property
    def httpClient(self) -> HTTPClient:
        """Http client"""
        if self._httpClient is None:
            raise NotInitialized
        return self._httpClient

    async def _saveImage(
        self, image: bytes, bucket: str, imgType: Literal["JPEG", "PNG"] | None = "JPEG", ratio: float | None = None
    ) -> str:
        """
        Save image to Luna-Image-Store
        Args:
            image: image as bytes
            bucket: bucket to save the image to
            imgType: image format
            ratio: image scale ratio
        Returns:
            link to saved image which will point to luna-api
        """
        headers = {}
        if ratio is not None:
            headers["X-Luna-Meta-Rescale"] = "{:.5f}".format(ratio)
        imageId = (
            await self.lunaImageStoreClient.postImage(
                bucketName=bucket,
                accountId=self.eventMetadata.accountId,
                contentType="image/jpeg" if imgType == "JPEG" else "image/png",
                imageInBytes=image,
                headers=headers,
                asyncRequest=True,
                raiseError=True,
            )
        ).json["image_id"]
        return f"{self.lunaApiExternalParams.origin}/{self._lunaApiExternalParams.apiVersion}/images/{imageId}"

    async def dumpEvent(self, lunaContext: LunaContext, aggregatedEvent: AggregatedEventT, eventType: str) -> dict:
        """
        Dump event using external metadata.
        Args:
            lunaContext: luna context
            aggregatedEvent: aggregated event made by analytics
            eventType: type of event made by analytics
        Returns:
            event enriched using external metadata
        """
        enrichedEvent = {
            "event_type": eventType,
            "account_id": self.eventMetadata.accountId,
            "event_create_time": self.eventMetadata.eventCreateTime,
            "event_end_time": self.eventMetadata.eventEndTime,
            "event": {
                **aggregatedEvent.asDict(),
                "location": self.eventMetadata.location,
                "stream_id": self.eventMetadata.streamId,
            },
        }
        if self.eventMetadata.source is not None:
            enrichedEvent["event"]["source"] = self.eventMetadata.source
        if self.eventMetadata.meta is not None:
            enrichedEvent["event"]["meta"] = self.eventMetadata.meta
        return enrichedEvent
